"""
torchsystem.loaders

This module contains some loader functions to create data loaders for the
training set and the validation set. If you want to add a new dataset
loader, just create a new function here. The loader function:
- must have an instance of the TorchExperiment as first input;
- can have several inputs; the TorchExperiment class will pass this inputs
  when calling the loader_fun using the loader_args tuple;
- must compute and return the followings:
    * train_loader: torch.utils.data.DataLoader instance for train set
    * valid_loader: torch.utils.data.DataLoader instance for validation set
    * test_loader: torch.utils.data.DataLoader instance for test set
    * input_shape: tuple 1 * 3
        where each element of this tuple means the following:
        (n_rows (height) * n_cols (width) * n_channels (depth))
    * classes: (only if classification experiment) tuple 1 * n_classes
Take a look to the MNIST dataset below for an example.
"""

import os
from torchvision import datasets
from torchvision import transforms
from pydaves.torchsystem.torchsystem import TorchExperiment
from torch.utils.data.dataloader import DataLoader


def mnist(experiment: TorchExperiment):
    # computing dataset path
    root = os.path.join(experiment.work_dir, "datasets")

    # computing transformations to do on dataset once it is loaded.
    # transforms.Compose: compose different transformations
    # transform.ToTensor: convert PIL Image or numpy.ndarray to tensor.
    # transform.Normalize: normalize a tensor image with mean and standard
    # deviation.
    # Given mean: (M1,...,Mn) and std: (S1,..,Sn) for n channels
    # this transform will normalize each channel of the input torch.*Tensor
    # i.e. input[channel] = (input[channel] - mean[channel]) / std[channel]
    transform = transforms.Compose(
        [transforms.ToTensor(),
         transforms.Normalize((0.1307,), (0.3081,))]
    )

    # getting training and validation set
    train_set = datasets.MNIST(
        root=root,
        train=True,
        download=True,
        transform=transform
    )
    test_set = datasets.MNIST(
        root=root,
        train=False,
        download=True,
        transform=transform
    )

    # crating loaders
    train_loader = DataLoader(
        train_set,
        batch_size=experiment.batch_size,
        shuffle=True
    )
    valid_loader = None
    test_loader = DataLoader(
        test_set,
        batch_size=experiment.batch_size,
        shuffle=False
    )

    # creating input_shape and classes
    input_shape = (28, 28, 1)
    classes = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9")

    # TODO: more smart way to pass outputs. Maybe the user does not want to
    #  use validation, or test also. Furthermore, input_shape and classes are
    #  not needed, cause the TorchExperiment instance can compute them using
    #  loaders during the init procedure.

    # returning loaders
    return train_loader, valid_loader, test_loader, input_shape, classes



import os
import torch
from pydaves.utils import PyConfig
from pydaves.utils import datetime_4_directory_string


def count_maxpool(block):
    dict_block = dict(block.named_children())
    n_maxpool = 0
    for key in dict_block.keys():
        # noinspection PyProtectedMember
        if 'MaxPool' in dict_block[key]._get_name():
            n_maxpool += 1
    return n_maxpool


def init_params(element):
    if type(element) in [torch.nn.Conv2d]:
        # initializing weights with kaiming normal distribution
        torch.nn.init.kaiming_normal_(element.weight)

        # initializing biases with normal distribution
        torch.nn.init.normal_(element.bias)


class TorchConfig(PyConfig):
    def __init__(self, file_path: str):
        # calling superclass __init__ method
        super().__init__(file_path=file_path)

        # pre-allocating experiment_settings
        self.__experiment_settings = None
        self.__classification_flag = None
        self.__reproducibility_flag = None
        self.__random_seed = None
        self.__verbose = None

        # pre-allocating path_settings
        self.__path_settings = None
        self.__main_dir = None
        self.__datetime_dir_name = datetime_4_directory_string()
        self.__version = None

        # pre-allocating logger_settings
        self.__logger_settings = None
        self.__console_only_flag = None
        self.__print_timestamp_flag = None

        # pre-allocating loader_settings
        self.__loader_settings = None
        self.__dataset_name = None

        # pre-allocating model_settings
        self.__model_settings = None
        self.__model_name = None

        # pre-allocating activation_settings
        self.__activation_settings = None
        self.__activation_name = None

        # pre-allocating loss_settings
        self.__loss_settings = None
        self.__loss_name = None

        # pre-allocating optimizer_settings
        self.__optimizer_settings = None
        self.__optimizer_name = None

        # pre-allocating lr_scheduler_settings
        self.__lr_scheduler_settings = None
        self.__lr_scheduler_name = None

        # pre-allocating training_settings
        self.__training_settings = None
        self.__n_epochs = None
        self.__batch_size = None
        self.__save_net_every_epochs = None
        self.__resume_flag = None
        self.__use_gpu_flag = None
        self.__gpu_index = None

        # pre-allocating testing_settings
        self.__testing_settings = None

        # setting default structure
        self.__default = {
            "experiment_settings": {
                "classification_flag": [None, bool],
                "reproducibility_flag": [False, bool],
                "random_seed": [2294, int],
                "verbose": [2, int]
                # TODO: add support to channel_first, channel_last
            },
            "path_settings": {
                "main_dir": [os.getcwd(), str],
                "version": ["MyTorchExperiment", str]
            },
            "logger_settings": {
                "console_only_flag": [False, bool],
                "print_timestamp_flag": [True, bool]
            },
            "loader_settings": {},
            "model_settings": {},
            "activation_settings": {
                "activation_name": [None, str]
            },
            "loss_settings": {
                "loss_name": [None, str]
            },
            "optimizer_settings": {
                "optimizer_name": [None, str]
            },
            "lr_scheduler_settings": {
                "lr_scheduler_name": [None, str]
            },
            "training_settings": {
                "n_epochs": [None, int],
                "batch_size": [None, int],
                "save_net_every_epochs": [None, int],
                "resume_flag": [False, bool],
                "use_gpu_flag": [True, bool],
                "gpu_index": [0, int]
            },
            "testing_settings": {}
        }

        # validating configuration
        self._validate_config()

    def __super_validator(self, attr_name, attr_value, attr_type):
        if isinstance(attr_value, attr_type):
            self.__setattr__(f"_TorchConfig__{attr_name}",
                             attr_value)
        else:
            raise TypeError(f"{attr_name} must be of type {attr_type}, "
                            f"got {type(attr_value)} instead")

    def __ismandatory(self, super_key, key=None):
        if key is None:
            # checking if the current super_key is mandatory
            mandatory_flag = False
            for c_key in self.__default[super_key]:
                if self.__ismandatory(super_key, c_key):
                    mandatory_flag = True
                    break
            return mandatory_flag
        elif self.__default[super_key][key][0] is None:
            # if the current default key is None, then an external defined one
            # is required, so the kay is mandatory
            return True
        else:
            return False

    def _validate_config(self):
        for super_key in self.__default:
            if super_key in self.data:
                for key in self.__default[super_key]:
                    if key in self.data[super_key]:
                        self.__super_validator(
                            attr_name=key,
                            attr_value=self.data[super_key][key],
                            attr_type=self.__default[super_key][key][1]
                        )
                    elif self.__ismandatory(super_key, key):
                        raise ValueError(f"Value {super_key}:{key} not found "
                                         "in the configuration file "
                                         f"{self.file_name}.\n"
                                         "            This value is "
                                         "mandatory, you must define it.")
                    else:
                        self.__super_validator(
                            attr_name=key,
                            attr_value=self.__default[super_key][key][0],
                            attr_type=self.__default[super_key][key][1]
                        )
                self.__setattr__(f"_TorchConfig__{super_key}",
                                 self.data[super_key])
            elif self.__ismandatory(super_key):
                raise ValueError(f"Section {super_key} not found "
                                 "in the configuration file "
                                 f"{self.file_name}.\n"
                                 "            This section is "
                                 "mandatory, you must define it.")
            else:
                self.__setattr__(f"_TorchConfig__{super_key}", {})
                for key in self.__default[super_key]:
                    self.__super_validator(
                        attr_name=key,
                        attr_value=self.__default[super_key][key][0],
                        attr_type=self.__default[super_key][key][1]
                    )

    @property
    def experiment_settings(self):
        return self.__experiment_settings

    @property
    def classification_flag(self):
        return self.__classification_flag

    @property
    def reproducibility_flag(self):
        return self.__reproducibility_flag

    @property
    def random_seed(self):
        return self.__random_seed

    @property
    def verbose(self):
        return self.__verbose

    @property
    def path_settings(self):
        return self.__path_settings

    @property
    def main_dir(self):
        return self.__main_dir

    @property
    def version(self):
        return self.__version

    @property
    def work_dir(self):
        return os.path.join(self.main_dir, self.version)

    @property
    def this_dir(self):
        return os.path.join(self.work_dir, self.__datetime_dir_name)

    @property
    def log_path(self):
        return os.path.join(self.this_dir, "log.bin")

    @property
    def logger_settings(self):
        return self.__logger_settings

    @property
    def console_only_flag(self):
        return self.__console_only_flag

    @property
    def print_timestamp_flag(self):
        return self.__print_timestamp_flag

    @property
    def loader_settings(self):
        return self.__loader_settings

    @property
    def model_settings(self):
        return self.__model_settings

    @property
    def activation_settings(self):
        return self.__activation_settings

    @property
    def activation_name(self):
        return self.__activation_name

    @property
    def loss_settings(self):
        return self.__loss_settings

    @property
    def loss_name(self):
        return self.__loss_name

    @property
    def optimizer_settings(self):
        return self.__optimizer_settings

    @property
    def optimizer_name(self):
        return self.__optimizer_name

    @property
    def lr_scheduler_settings(self):
        return self.__lr_scheduler_settings

    @property
    def lr_scheduler_name(self):
        return self.__lr_scheduler_name

    @property
    def training_settings(self):
        return self.__training_settings

    @property
    def n_epochs(self):
        return self.__n_epochs

    @property
    def batch_size(self):
        return self.__batch_size

    @property
    def save_net_every_epochs(self):
        return self.__save_net_every_epochs

    @property
    def resume_flag(self):
        return self.__resume_flag

    @property
    def use_gpu_flag(self):
        return self.__use_gpu_flag

    @property
    def gpu_index(self):
        return self.__gpu_index

    @property
    def testing_settings(self):
        return self.__testing_settings

    def __repr__(self):
        return f"TorchConfig instance pointing to {self.file_name}"

    def __str__(self):
        return self.__repr__()


class R2Score(object):
    """
        Calculates the R-Squared, the coefficient of determination
        <https://en.wikipedia.org/wiki/Coefficient_of_determination>

        - update method must receive output of the form (y_pred, y_true) or
        {'y_pred': y_pred, 'y_true': y_true}.
        - y and y_pred must be of same shape (N, ) or (N, 1) and of type
    """

    # FIXME: does not work (returns values > 1)

    def __init__(self):
        self._num_examples = 0
        self._sum_of_errors = 0
        self._y_sq_sum = 0
        self._y_sum = 0

    def reset(self):
        self._num_examples = 0
        self._sum_of_errors = 0
        self._y_sq_sum = 0
        self._y_sum = 0

    def update(self, output):
        y_pred, y_true = output
        self._num_examples += y_true.shape[0]
        self._sum_of_errors += torch.sum(torch.pow(y_pred - y_true, 2)).item()

        self._y_sum += torch.sum(y_true).item()
        self._y_sq_sum += torch.sum(torch.pow(y_true, 2)).item()

    def compute(self):
        if self._num_examples == 0:
            raise RuntimeError('R2Score must have at least one example '
                               'before it can be computed.')
        return 1 - self._sum_of_errors / (
                self._y_sq_sum - (self._y_sum ** 2) / self._num_examples)


if __name__ == "__main__":
    config = TorchConfig("config_example.yml")
    print(config)



import os
import datetime


def touch_dir(dir_path):
    if not (os.path.exists(dir_path)):
        os.makedirs(dir_path)


def datetime_4_directory_string():
    now = datetime.datetime.now()
    return "{}-{}-{}_{}-{}-{}".format(
        str(now.year).zfill(4),
        str(now.month).zfill(2),
        str(now.day).zfill(2),
        str(now.hour).zfill(2),
        str(now.minute).zfill(2),
        str(now.second).zfill(2))


def time_string():
    now = datetime.datetime.now()
    return "{}:{}:{}".format(
        str(now.hour).zfill(2),
        str(now.minute).zfill(2),
        str(now.second).zfill(2))


def date_string():
    now = datetime.datetime.now()
    return "{}-{}-{}".format(
        str(now.year).zfill(4),
        str(now.month).zfill(2),
        str(now.day).zfill(2))


def datetime_string():
    return f"{date_string()} {time_string()}"


def set_fontsize(ax, fontsize: int):
    for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                 ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(fontsize)


def scatter_plot_colored(df, x, y, c='', ax=None, diagonal=True, fontsize=20):
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    # checking if an ax is passed
    if ax is None:
        _, ax = plt.subplots(figsize=(20, 15))

    # checking if a color is passed
    if c == "":
        # plotting on ax without color
        ax.scatter(df[x].values, df[y])
    else:
        # getting color_map
        # noinspection SpellCheckingInspection,PyUnresolvedReferences
        color_map = mpl.cm.get_cmap('coolwarm')

        # getting color values
        z = df[c].values

        # normalizing mpl.colors on z min and max
        # noinspection PyUnresolvedReferences
        normalize = mpl.colors.Normalize(vmin=min(z), vmax=max(z))

        # computing colors
        colors = [color_map(normalize(value)) for value in z]

        # plotting on ax
        ax.scatter(df[x].values, df[y], color=colors)

        # setting color bar
        # noinspection PyUnresolvedReferences
        color_bar_ax, _ = mpl.colorbar.make_axes(ax, location="right")

        # noinspection PyUnresolvedReferences
        color_bar = mpl.colorbar.ColorbarBase(
            colorbar_ax, cmap=color_map, norm=normalize)
        color_bar.set_label(c, rotation=90, fontsize=fontsize)
        set_fontsize(color_bar_ax, fontsize)

    # setting axis properties
    ax.set_facecolor('white')
    ax.grid(linestyle=':', linewidth='0.5', color='grey')
    ax.set_xlabel(x)
    ax.set_ylabel(y)

    # plotting diagonal if needed
    if diagonal:
        lb = df[[x, y]].min().min() - .1
        ub = df[[x, y]].max().max() + .1
        ax.plot([lb, ub], [lb, ub], 'k--')

    # modifying all font sizes
    set_fontsize(ax, fontsize)

    # returning the ax right now created
    return ax


# noinspection PyPep8Naming
def train_test_valid_split(X, y, test_size=0.2, valid_size=0.1,
                           random_state=2294, shuffle=False):
    # TODO: recomputing test_size e valid_size in order to have right
    #  splitting, instead a validation set stealing data from test
    import numpy as np
    import pandas as pd
    from sklearn.model_selection import train_test_split
    # creating an empty DataFrame for X and y
    if isinstance(X, pd.DataFrame):
        X_empty = pd.DataFrame([], columns=X.columns.to_list())
        y_empty = pd.DataFrame([], columns=y.columns.to_list())
    else:
        X_empty = np.array([])
        y_empty = np.array([])

    # pre-allocating splits
    X_test = X_empty
    y_test = y_empty
    X_valid = X_empty
    y_valid = y_empty

    # checking test_frac
    if test_size > .0:
        X_train, X_test, y_train, y_test = train_test_split(
            X, y,
            test_size=test_size,
            random_state=random_state,
            shuffle=shuffle)

        # checking valid_frac
        if valid_size > .0:
            X_test, X_valid, y_test, y_valid = train_test_split(
                X_test, y_test,
                test_size=valid_size,
                random_state=random_state,
                shuffle=shuffle)
    else:
        X_train = X
        y_train = y

    # returning splits
    return X_train, X_test, X_valid, y_train, y_test, y_valid


class PyLogger(object):
    def __init__(self, log_path="log.bin",
                 console_only_flag=False, print_timestamp_flag=True):
        # validating inputs
        self.__log_path, self.console_only_flag, self.print_timestamp_flag = \
            self._validate_inputs(
                log_path,
                console_only_flag, print_timestamp_flag)

        # pre-allocating the log stream
        self.__log_stream = None

        # starting the logging procedure
        self.start()

    def __repr__(self):
        return f"PyLogger instance pointing to {self.log_name}"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def _validate_inputs(log_path: str,
                         console_only_flag: bool, print_timestamp_flag: bool):
        # checking log_path data type and creating the entire path if needed
        if not isinstance(log_path, str):
            raise TypeError("Input 'log_path' must be of type 'str', got "
                            f"{type(log_path)} instead.")
        elif os.path.split(log_path)[0] == "":
            log_path = os.path.join(os.getcwd(), log_path)
        if not isinstance(console_only_flag, bool):
            raise TypeError("Input 'console_only_flag' must be of type "
                            f"'bool', got {type(console_only_flag)} instead.")
        if not isinstance(print_timestamp_flag, bool):
            raise TypeError("Input 'print_timestamp_bool' must be of type "
                            f"'bool', got {type(print_timestamp_flag)} "
                            f"instead.")
        return log_path, console_only_flag, print_timestamp_flag

    @property
    def file_path(self):
        return self.__log_path

    @property
    def file_dir(self):
        return os.path.split(self.file_path)[0]

    @property
    def file_name(self):
        return os.path.split(self.file_path)[1]

    @property
    def file_ext(self):
        return os.path.splitext(self.file_name)[1]

    @property
    def log_path(self):
        return self.__log_path

    @property
    def log_dir(self):
        return self.file_dir

    @property
    def log_name(self):
        return self.file_name

    @property
    def log_ext(self):
        return self.file_ext

    def start(self):
        # creating and / or opening a new log file
        if not self.console_only_flag:
            self.__log_stream = open(self.log_path, "a")

    def stop(self):
        # closing the log file
        if not self.console_only_flag:
            self.__log_stream.close()

    def _check_status(self):
        # opening the log file if necessary
        if self.__log_stream is None:
            self.start()
        elif self.__log_stream.closed:
            self.start()

    def print(self,
              inp_text: str,
              style="",
              up_extra_lines=0,
              dw_extra_lines=0,
              timestamp_flag=None):
        """
        :param inp_text: input text to be shown
        :param style: specify the dash style; it can be:
            "" (normal, no dash at all);
            "u" (upper dashed);
            "d" (down dashed);
            "ud" (upper and down dashed)
        :param up_extra_lines: number of empty lines above the print
        :param dw_extra_lines: number of empty lines below the print
        :param timestamp_flag: specify if the timestamp has to be printed
        """
        # checking the log file status
        if not self.console_only_flag:
            self._check_status()

        # adding the timestamp to input_text if required
        if timestamp_flag is None:
            timestamp_flag = self.print_timestamp_flag
        if timestamp_flag:
            inp_text = time_string() + " | " + inp_text

        # pre-allocating output text
        out_text = "\n" * up_extra_lines

        # writing output text
        if style.__contains__("u"):
            out_text += self.dash(inp_text) + "\n"
        out_text += inp_text + "\n"
        if style.__contains__("d"):
            out_text += self.dash(inp_text) + "\n"

        # printing extra lines
        out_text += "\n" * dw_extra_lines

        # printing on console
        print(out_text, end="")

        # printing on log file
        if not self.console_only_flag:
            self.__log_stream.write(out_text)

    @staticmethod
    def dash(text):
        return "=" * len(text)


class PyConfig(object):
    def __init__(self, file_path="config.yml"):
        # validating inputs
        self.__file_path = self._validate_inputs(file_path)

        # loading routine
        self._data = self._load()

    def __repr__(self):
        return f"PyConfig instance pointing to {self.file_name}"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def _validate_inputs(file_path: str):
        # checking file_path data type and creating the entire path if needed
        if not isinstance(file_path, str):
            raise TypeError("Input must be of type 'str'")
        elif os.path.split(file_path)[0] == "":
            file_path = os.path.join(os.getcwd(), file_path)

        # checking file existence
        try:
            open(file_path)
        except FileExistsError:
            raise FileNotFoundError(f"Input file {file_path} does not exist")

        # returning validated input
        return file_path

    def _load(self):
        import yaml
        # checking for the config file extension
        if self.file_ext in [".yml", ".yaml"]:
            # loading as yaml file
            data = yaml.safe_load(open(self.__file_path))
        elif self.file_ext == ".json":
            # loading as json file
            # TODO: create a loading routine for json files
            data = None
        elif self.file_ext in [".ini", ".bin", ".txt"]:
            # loading with the configparser
            # TODO: create a loading routine with the configparser
            data = None
        else:
            # raising exception for unrecognized extension
            raise ValueError(f"Unrecognized extension: {self.file_ext}")
        # returning data
        return data

    @property
    def file_path(self):
        return self.__file_path

    @property
    def file_dir(self):
        return os.path.split(self.__file_path)[0]

    @property
    def file_name(self):
        return os.path.split(self.__file_path)[1]

    @property
    def file_ext(self):
        return os.path.splitext(self.file_name)[1]

    @property
    def data(self):
        return self._data

    def reset(self):
        self.__init__(self.file_path)

    def reset_inputs(self, file_path="config.yml"):
        self.__init__(file_path)

    def write(self):
        # TODO: writing routine for all file types
        pass


class PGConnectionManager(object):
    def __init__(self, dbname, user, password, host, port):
        self.dbname = dbname
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.connection = None
        self.cursor = None

    def __enter__(self):
        import psycopg2
        self.connection = psycopg2.connect(
            dbname=self.dbname,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port
        )
        self.cursor = self.connection.cursor()
        return self.connection, self.cursor

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cursor.close()
        self.connection.close()


# noinspection SqlNoDataSourceInspection
def pg_exists_table(cursor, table_name: str, schema_name="public") -> bool:
    query = f"SELECT EXISTS (\n" \
            f"    SELECT FROM information_schema.tables\n" \
            f"    WHERE table_schema = '{schema_name}'\n" \
            f"    AND   table_name = '{table_name}'\n" \
            f"    );"
    cursor.execute(query)
    return cursor.fetchone()[0]



import os
import torch
import torchnet
import traceback
import numpy as np
from pydaves.utils import PyLogger
from pydaves.utils import touch_dir
from pydaves.torchsystem.torchutils import R2Score
from pydaves.torchsystem.torchutils import TorchConfig
from torch.utils.data.dataloader import DataLoader


class TorchExperiment(object):
    def __init__(self,
                 config: TorchConfig,
                 loader_fun, model_class,
                 loader_args=(), model_args=()):
        # validating inputs
        self.__config, \
        self.__loader_fun, self.__model_class, \
        self.__loader_args, self.__model_args = \
            self._validate_inputs(
                config,
                loader_fun, model_class,
                loader_args, model_args
            )

        # managing reproducibility
        if self.reproducibility_flag:
            # torch.manual_seed(self.random_seed)
            # np.random.seed(self.random_seed)
            # FIXME: modifying random_seed means weird results in training
            pass

        # managing paths
        touch_dir(self.work_dir)
        touch_dir(self.this_dir)

        # creating a logger instance
        self.__logger = PyLogger(
            log_path=self.log_path,
            console_only_flag=self.console_only_flag,
            print_timestamp_flag=self.print_timestamp_flag
        )

        # creating a device instance
        self.__device = self._validate_gpu()

        # creating loaders
        if self.classification_flag:
            self.__train_loader, self.__valid_loader, self.__test_loader, \
            self.__input_shape, self.__classes = self.loader_fun(
                self, *loader_args)
        else:
            self.__train_loader, self.__valid_loader, self.__test_loader, \
            self.__input_shape = self.loader_fun(
                self, *loader_args)
            self.__classes = ()

        # creating activation and loss
        self.__activation = self._get_torch_activation()
        self.__loss = self._get_torch_loss()

        # creating a model instance
        self._model = self.model_class(
            self, *model_args)

        # creating optimizer and lr_scheduler
        self.__optimizer = self._get_torch_optimizer()
        self.__lr_scheduler = self._get_torch_lr_scheduler()

        # creating a trainer instance
        self.__trainer = _TorchTrainer(experiment=self)

    @staticmethod
    def _validate_inputs(config: TorchConfig,
                         loader_fun, model_class,
                         loader_args: tuple, model_args: tuple):
        if not isinstance(config, TorchConfig):
            raise TypeError("Input 'config' must be of type 'TorchConfig', "
                            f"got {type(config)} instead.")
        if not callable(loader_fun):
            raise TypeError("Input 'loader_fun' must be callable.")
        if not callable(model_class):
            raise TypeError("Input 'model_class' must be callable.")
        if not isinstance(loader_args, tuple):
            raise TypeError("Input 'loader_args' must be of type 'tuple', "
                            f"got {type(loader_args)} instead.")
        if not isinstance(model_args, tuple):
            raise TypeError("Input 'model_args' must be of type 'tuple', "
                            f"got {type(model_args)} instead.")
        return config, loader_fun, model_class, loader_args, model_args

    def _validate_gpu(self):
        # checking if the selected gpu is available
        if self.use_gpu_flag:
            if torch.cuda.is_available():
                if self.gpu_index < torch.cuda.device_count():
                    # enabling the chosen gpu
                    os.environ["CUDA_VISIBLE_DEVICES"] = str(self.gpu_index)

                    # creating a torch.device instance for the chosen gpu
                    device = torch.device(f"cuda:{self.gpu_index}")
                else:
                    raise ValueError(f"Invalid gpu index ({self.gpu_index}).")
            else:
                raise ValueError("CUDA not available.")
        else:
            # creating device for cpu
            device = torch.device("cpu")
        return device

    def _get_torch_activation(self):
        # getting activation_name
        activation_name = self.activation_name.lower()

        # selecting activation
        # TODO: add here a routine for each Torch activation
        if activation_name == "tanh":
            return torch.nn.Tanh()
        elif activation_name == "elu":
            if "alpha" not in self.activation_settings:
                self.set(
                    keys=("activation_settings", "alpha"),
                    value=1.0
                )
            if "inplace" not in self.activation_settings:
                self.set(
                    keys=("activation_settings", "inplace"),
                    value=False
                )
            return torch.nn.ELU(
                alpha=self.activation_settings["alpha"],
                inplace=self.activation_settings["inplace"]
            )
        else:
            raise ValueError(f"Unrecognized activation_name "
                             f"{self.activation_name}.")

    def _get_torch_loss(self):
        # getting loss_name
        loss_name = self.loss_name.lower()

        # selecting loss
        # TODO: add here a routine for each Torch loss
        if loss_name == "crossentropyloss":
            if "ignore_index" not in self.loss_settings:
                self.set(
                    keys=("loss_settings", "ignore_index"),
                    value=-100
                )
            if "reduction" not in self.loss_settings:
                self.set(
                    keys=("loss_settings", "reduction"),
                    value="mean"
                )
            return torch.nn.CrossEntropyLoss(
                ignore_index=self.loss_settings["ignore_index"],
                reduction=self.loss_settings["reduction"]
            )
        elif loss_name == "mseloss":
            if "reduction" not in self.loss_settings:
                self.set(
                    keys=("loss_settings", "reduction"),
                    value="mean"
                )
            return torch.nn.MSELoss(
                reduction=self.loss_settings["reduction"]
            )
        else:
            raise ValueError(f"Unrecognized loss_name "
                             f"{self.loss_name}.")

    # noinspection DuplicatedCode
    def _get_torch_optimizer(self):
        # getting optimizer_name
        optimizer_name = self.optimizer_name.lower()

        # selecting optimizer
        # TODO: add here a routine for each Torch optimizer
        if optimizer_name == "adam":
            if "lr" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "lr"),
                    value=0.001
                )
            if "betas" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "betas"),
                    value=(0.9, 0.999)
                )
            if "eps" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "eps"),
                    value=1e-08
                )
            if "weight_decay" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "weight_decay"),
                    value=0
                )
            if "amsgrad" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "amsgrad"),
                    value=False
                )
            return torch.optim.Adam(
                params=self.model.parameters(),
                lr=self.optimizer_settings["lr"],
                betas=self.optimizer_settings["betas"],
                eps=self.optimizer_settings["eps"],
                weight_decay=self.optimizer_settings["weight_decay"],
                amsgrad=self.optimizer_settings["amsgrad"]
            )
        elif optimizer_name == "adamax":
            if "lr" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "lr"),
                    value=0.002
                )
            if "betas" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "betas"),
                    value=(0.9, 0.999)
                )
            if "eps" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "eps"),
                    value=1e-08
                )
            if "weight_decay" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "weight_decay"),
                    value=0
                )
            # noinspection PyUnresolvedReferences
            return torch.optim.Adamax(
                params=self.model.parameters(),
                lr=self.optimizer_settings["lr"],
                betas=self.optimizer_settings["betas"],
                eps=self.optimizer_settings["eps"],
                weight_decay=self.optimizer_settings["weight_decay"],
            )
        elif optimizer_name == "sgd":
            if "lr" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "lr"),
                    value=0.01
                )
            if "momentum" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "momentum"),
                    value=0
                )
            if "dampening" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "dampening"),
                    value=0
                )
            if "weight_decay" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "weight_decay"),
                    value=0
                )
            if "nesterov" not in self.optimizer_settings:
                self.set(
                    keys=("optimizer_settings", "nesterov"),
                    value=False
                )
            return torch.optim.SGD(
                params=self.model.parameters(),
                lr=self.optimizer_settings["lr"],
                momentum=self.optimizer_settings["momentum"],
                dampening=self.optimizer_settings["dampening"],
                weight_decay=self.optimizer_settings["weight_decay"],
                nesterov=self.optimizer_settings["nesterov"]
            )
        else:
            raise ValueError(f"Unrecognized optimizer_name "
                             f"{self.optimizer_name}.")

    def _get_torch_lr_scheduler(self):
        # getting lr_scheduler_name
        lr_scheduler_name = self.lr_scheduler_name.lower()

        # selecting lr_scheduler
        # TODO: add here a routine for each Torch lr_scheduler
        # noinspection SpellCheckingInspection
        if lr_scheduler_name == "reducelronplateau":
            if "mode" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "mode"),
                    value="min"
                )
            if "factor" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "factor"),
                    value=0.1
                )
            if "patience" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "patience"),
                    value=10
                )
            if "verbose" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "verbose"),
                    value=False
                )
            if "threshold" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "threshold"),
                    value=1e-4
                )
            if "threshold_mode" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "threshold_mode"),
                    value="rel"
                )
            if "cooldown" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "cooldown"),
                    value=0
                )
            if "min_lr" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "min_lr"),
                    value=0
                )
            if "eps" not in self.lr_scheduler_settings:
                self.set(
                    keys=("lr_scheduler_settings", "eps"),
                    value=1e-8
                )
            return torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer=self.optimizer,
                mode=self.lr_scheduler_settings["mode"],
                factor=self.lr_scheduler_settings["factor"],
                patience=self.lr_scheduler_settings["patience"],
                verbose=self.lr_scheduler_settings["verbose"],
                threshold=self.lr_scheduler_settings["threshold"],
                threshold_mode=self.lr_scheduler_settings["threshold_mode"],
                cooldown=self.lr_scheduler_settings["cooldown"],
                min_lr=self.lr_scheduler_settings["min_lr"],
                eps=self.lr_scheduler_settings["eps"]
            )
        else:
            raise ValueError(
                "<< {} >> lr_scheduler is not a valid one.".format(
                    self.lr_scheduler))

    @property
    def config(self) -> TorchConfig:
        return self.__config

    @property
    def loader_fun(self):
        return self.__loader_fun

    @property
    def model_class(self):
        return self.__model_class

    @property
    def loader_args(self) -> tuple:
        return self.__loader_args

    @property
    def model_args(self) -> tuple:
        return self.__model_args

    @property
    def logger(self) -> PyLogger:
        return self.__logger

    @property
    def device(self) -> torch.device:
        return self.__device

    @property
    def train_loader(self) -> DataLoader:
        return self.__train_loader

    @property
    def valid_loader(self) -> DataLoader:
        return self.__valid_loader

    @property
    def test_loader(self) -> DataLoader:
        return self.__test_loader

    # noinspection PyUnresolvedReferences
    @property
    def n_batches_train(self) -> int:
        return len(self.train_loader)

    @property
    def n_batches_valid(self) -> int:
        if self.valid_loader is None:
            return 0
        else:
            return len(self.valid_loader)

    @property
    def n_batches_test(self) -> int:
        return len(self.test_loader)

    @property
    def input_shape(self) -> tuple:
        return self.__input_shape

    @property
    def input_height(self) -> int:
        return self.input_shape[0]

    @property
    def input_width(self) -> int:
        return self.input_shape[1]

    @property
    def input_depth(self) -> int:
        return self.input_shape[2]

    @property
    def input_n_rows(self) -> int:
        return self.input_shape[0]

    @property
    def input_n_cols(self) -> int:
        return self.input_shape[1]

    @property
    def input_n_channels(self) -> int:
        return self.input_shape[2]

    @property
    def classes(self) -> tuple:
        return self.__classes

    @property
    def n_classes(self) -> int:
        return len(self.classes)

    @property
    def model(self):
        return self._model

    @property
    def model_name(self) -> str:
        return type(self.model).__name__

    @property
    def activation(self):
        return self.__activation

    @property
    def loss(self):
        return self.__loss

    @property
    def optimizer(self):
        return self.__optimizer

    @property
    def lr_scheduler(self):
        return self.__lr_scheduler

    @property
    def trainer(self):
        return self.__trainer

    @property
    def experiment_settings(self) -> dict:
        return self.__config.experiment_settings

    @property
    def classification_flag(self) -> bool:
        return self.__config.classification_flag

    @property
    def reproducibility_flag(self) -> bool:
        return self.__config.reproducibility_flag

    @property
    def random_seed(self) -> int:
        return self.__config.random_seed

    @property
    def verbose(self) -> int:
        return self.__config.verbose

    @property
    def path_settings(self) -> dict:
        return self.__config.path_settings

    @property
    def main_dir(self) -> str:
        return self.__config.main_dir

    @property
    def version(self) -> str:
        return self.__config.version

    @property
    def work_dir(self) -> str:
        return self.__config.work_dir

    @property
    def this_dir(self) -> str:
        return self.__config.this_dir

    @property
    def log_path(self) -> str:
        return self.__config.log_path

    @property
    def logger_settings(self) -> dict:
        return self.__config.logger_settings

    @property
    def console_only_flag(self) -> bool:
        return self.__config.console_only_flag

    @property
    def print_timestamp_flag(self) -> bool:
        return self.__config.print_timestamp_flag

    @property
    def loader_settings(self) -> dict:
        return self.__config.loader_settings

    @property
    def model_settings(self):
        return self.__config.model_settings

    @property
    def activation_settings(self):
        return self.__config.activation_settings

    @property
    def activation_name(self) -> str:
        return self.__config.activation_name

    @property
    def loss_settings(self) -> dict:
        return self.__config.loss_settings

    @property
    def loss_name(self) -> str:
        return self.__config.loss_name

    @property
    def optimizer_settings(self) -> dict:
        return self.__config.optimizer_settings

    @property
    def optimizer_name(self) -> str:
        return self.__config.optimizer_name

    @property
    def lr_scheduler_settings(self) -> dict:
        return self.__config.lr_scheduler_settings

    @property
    def lr_scheduler_name(self) -> str:
        return self.__config.lr_scheduler_name

    @property
    def training_settings(self) -> dict:
        return self.__config.training_settings

    @property
    def n_epochs(self) -> int:
        return self.__config.n_epochs

    @property
    def batch_size(self) -> int:
        return self.__config.batch_size

    @property
    def save_net_every_epochs(self) -> int:
        return self.__config.save_net_every_epochs

    @property
    def resume_flag(self) -> bool:
        return self.__config.resume_flag

    @property
    def use_gpu_flag(self) -> bool:
        return self.__config.use_gpu_flag

    @property
    def gpu_index(self) -> int:
        return self.__config.gpu_index

    @property
    def testing_settings(self) -> dict:
        return self.__config.testing_settings

    # noinspection PyProtectedMember
    def set(self, keys, value):
        if not isinstance(keys, tuple):
            raise TypeError("Input 'keys' must be of type 'tuple', "
                            f"got {type(keys)} instead.")
        elif len(keys) != 2:
            raise ValueError("Lenght of input 'keys' must be 2.")

        # checking the super_key existance
        if keys[0] not in self.config.data:
            self.config._data[keys[0]] = {}

        # assigning the key value
        self.config._data[keys[0]][keys[1]] = value

        # running again _validating_config routine
        self.config._validate_config()

    def print_summary(self):
        # summary preamble
        self.logger.print("EXPERIMENT SETTINGS:", style="d")

        # getting properties list
        prop_list = dir(self)

        # removing some properties from the list
        prop_list.remove("print_summary")
        prop_list.remove("run")
        prop_list.remove("set")

        for current_prop in prop_list:
            if current_prop[0] is not "_":
                self.logger.print("{} = {}".format(
                    current_prop, getattr(self, current_prop)),
                    timestamp_flag=False)

    def run(self):
        # the following try clause contains all the algorithm
        # noinspection PyBroadException
        try:
            if self.verbose >= 1:
                # noinspection SpellCheckingInspection
                self.logger.print(
                    "TORCHEXPERIMENT STARTED",
                    style="ud",
                    timestamp_flag=False
                )

            if self.verbose >= 2:
                self.print_summary()

            # sending model to the right device
            self.model.to(self.device)

            # training
            self.trainer.train()

            # test
            self.trainer.test()
        except Exception:
            self.logger.print(traceback.format_exc())

        # stopping the logger object
        self.logger.stop()

    def __repr__(self):
        return f"TorchExperiment to train {self.model_name}"


class _TorchTrainer(object):
    """
    Private Class, only TorchExperiment can use it.
    """

    def __init__(self, experiment: TorchExperiment):
        # assigning experiment
        self.experiment = experiment

        # pre-allocating result values
        if self.experiment.classification_flag:
            self.train_results = {
                "loss_list": [],
                "acc_list": [],
                "cm_list": [],
                "last_loss": float("inf"),
                "last_acc": 0.,
                "best_loss": float("inf"),
                "best_acc": 0.,
                "best_epoch_for_loss": 0,
                "best_epoch_for_acc": 0
            }
            self.valid_results = {
                "loss_list": [],
                "acc_list": [],
                "cm_list": [],
                "lr_list": [],
                "last_loss": float("inf"),
                "last_acc": 0.,
                "best_loss": float("inf"),
                "best_acc": 0.,
                "best_epoch_for_loss": 0,
                "best_epoch_for_acc": 0
            }
            self.test_results = {
                "loss": None,
                "acc": None,
                "cm": None
            }
        else:
            self.train_results = {
                "loss_list": [],
                "r2_list": [],
                "last_loss": float("inf"),
                "last_r2": 0.,
                "best_loss": float("inf"),
                "best_r2": 0.,
                "best_epoch_for_loss": 0,
                "best_epoch_for_r2": 0
            }
            self.valid_results = {
                "loss_list": [],
                "r2_list": [],
                "lr_list": [],
                "last_loss": float("inf"),
                "last_r2": 0.,
                "best_loss": float("inf"),
                "best_r2": 0.,
                "best_epoch_for_loss": 0,
                "best_epoch_for_r2": 0
            }
            self.test_results = {
                "loss": None,
                "r2": None
            }

    def __repr__(self):
        return f"TorchTrainer instance to train {self.experiment.model_name}"

    def __str__(self):
        return self.__repr__()

    def _init_training(self):
        self.__init__(self.experiment)

    def run(self):
        self.train()
        self.test()

    def train(self):
        # initializing training
        self._init_training()

        # main loop
        for epoch in range(self.experiment.n_epochs):
            # printing this epoch
            self.experiment.logger.print(
                'EPOCH {}/{}'.format(epoch + 1, self.experiment.n_epochs),
                style="ud", up_extra_lines=1)

            # running training
            self._loop(mode=0)

            if self.experiment.n_batches_valid > 0:
                # running validation
                self._loop(mode=1)

            # saving last values if best training loss
            if self.train_results["last_loss"] >= \
                    self.train_results["best_loss"]:
                self.train_results["best_loss"] = \
                    self.train_results["last_loss"]
                self.train_results["best_epoch_for_loss"] = epoch

            if self.experiment.classification_flag:
                # saving last values if best training acc
                if self.train_results["last_acc"] >= \
                        self.train_results["best_acc"]:
                    self.train_results["best_acc"] = \
                        self.train_results["last_acc"]
                    self.train_results["best_epoch_for_acc"] = epoch
            else:
                # saving last values if best training r2
                if self.train_results["last_r2"] >= \
                        self.train_results["best_r2"]:
                    self.train_results["best_r2"] = \
                        self.train_results["last_r2"]
                    self.train_results["best_epoch_for_r2"] = epoch

            # saving last net
            torch.save(
                self.experiment.model.state_dict(),
                os.path.join(self.experiment.this_dir,
                             'Net_last_update.pt')
            )

            # saving net if the epoch x has come
            if (epoch + 1) % self.experiment.save_net_every_epochs == 0:
                torch.save(
                    self.experiment.model.state_dict(),
                    os.path.join(self.experiment.this_dir,
                                 f"Net_at_epoch_{epoch + 1}.pt")
                )

            # saving last net and values if best validation loss
            if self.valid_results["last_loss"] <= \
                    self.valid_results["best_loss"]:
                torch.save(
                    self.experiment.model.state_dict(),
                    os.path.join(self.experiment.this_dir,
                                 'Net_best_loss.pt')
                )
                self.valid_results["best_loss"] = \
                    self.valid_results["last_loss"]
                self.valid_results["best_epoch_for_loss"] = epoch

            if self.experiment.classification_flag:
                # saving last net and values if best validation acc
                if self.valid_results["last_acc"] >= \
                        self.valid_results["best_acc"]:
                    torch.save(
                        self.experiment.model.state_dict(),
                        os.path.join(self.experiment.this_dir,
                                     'Net_best_acc.pt')
                    )
                    self.valid_results["best_acc"] = \
                        self.valid_results["last_acc"]
                    self.valid_results["best_epoch_for_acc"] = epoch
            else:
                # saving last net and values if best validation r2
                if abs(1 - self.valid_results["last_r2"]) <= \
                        abs(1 - self.valid_results["best_r2"]):
                    torch.save(
                        self.experiment.model.state_dict(),
                        os.path.join(self.experiment.this_dir,
                                     'Net_best_r2.pt')
                    )
                    self.valid_results["best_r2"] = \
                        self.valid_results["last_r2"]
                    self.valid_results["best_epoch_for_r2"] = epoch

            # TODO: print a summary for this epoch

            # TODO: print if a new best net was found

            # TODO: saving routine for plots

            # saving plot
            # self.save_loss_fig(losses_train, 'loss_val',
            # losses_val=losses_val)
            # self.save_loss_fig(accs_train, 'acc_val', losses_val=accs_val)
            # self.save_loss_fig(lr_epoch, 'learning_rate')

            # self.experiment.logger("")
            # print('== TRAIN ==')
            # print('LOSS: {:.4f}'.format(loss_tr))
            # print('ACC:  {:.4f}'.format(acc_tr))
            # print('CM: \n%s' % cm_tr)
            # print('== VALIDATION ==')
            # print('LOSS: {:.4f}'.format(loss_val))
            # print('ACC:  {:.4f}'.format(acc_val))
            # print('CM: \n%s' % cm_val)
            # print(f'Epoch best loss: {epoch_best}')
            # print(f'Epoch best acc:  {epoch_best_acc}')
            # print('\n')

    # noinspection PyArgumentList
    def _loop(self, mode):
        """
        Private method _loop
        :param mode: 0 for train, 1 for valid, 2 for test
        :return:
        """

        # switch mode
        if mode == 0:  # training
            if self.experiment.verbose > 0:
                self.experiment.logger.print(
                    "TRAINING", style="ud", up_extra_lines=1)
            # turning the model to train mode
            self.experiment.model.train()
            dataset_loader = self.experiment.train_loader
            n_batches = self.experiment.n_batches_train
            mode_str = "TRAINING"
            #
        elif mode == 1:  # validation
            if self.experiment.verbose > 0:
                self.experiment.logger.print(
                    "VALIDATION", style="ud", up_extra_lines=1)
            # turning the model to evaluation mode
            self.experiment.model.eval()
            dataset_loader = self.experiment.valid_loader
            n_batches = self.experiment.n_batches_valid
            mode_str = "VALIDATION"
            #
        elif mode == 2:  # testing
            if self.experiment.verbose > 0:
                self.experiment.logger.print(
                    "TESTING", style="ud", up_extra_lines=1)
            # turning the model to evaluation mode
            self.experiment.model.eval()
            dataset_loader = self.experiment.test_loader
            n_batches = self.experiment.n_batches_test
            mode_str = "TESTING"
            #
        else:
            raise ValueError(f"Unrecognized 'mode' value ({mode}), "
                             f"only 0/1/2 permitted.")

        # pre-allocating batch_loss_list
        batch_loss_list = []

        # creating instances to track accuracy / R2 score evolution
        if self.experiment.classification_flag:
            # confusion matrix if classification
            confusion_matrix = torchnet.meter.ConfusionMeter(
                self.experiment.n_classes)
        else:
            # R2 score if regression
            r2_score = R2Score()

        # batch loop
        for batch_index, batch_samples in enumerate(dataset_loader):
            # getting current X (inputs) and y_true (labels)
            X = batch_samples[0].float()
            y_true = batch_samples[1]

            # sending inputs and labels to the right device
            X = X.to(self.experiment.device)
            y_true = y_true.to(self.experiment.device)

            # clearing gradient buffer (we do not want any gradient from
            # previous epoch, we do not want to cumulate gradients)
            self.experiment.optimizer.zero_grad()

            # computing output
            y_pred = self.experiment.model(X)

            # deleting inputs from memory (if on gpu, releasing memory)
            del X

            # computing loss
            if self.experiment.loss_name.lower() == "mseloss":
                batch_loss = self.experiment.loss(
                    y_pred.float(), y_true.float()
                )
            else:
                batch_loss = self.experiment.loss(
                    y_pred, y_true
                )

            # updating batch loss list
            # noinspection PyTypeChecker
            batch_loss_list.append(float(batch_loss))

            if self.experiment.classification_flag:
                # computing class_pred
                _, classes_pred = torch.max(y_pred, 1)
                classes_true = y_true

                # now, both class_true and class_pred are 1 x batch_size
                # tensors. So, we can compute accuracy
                batch_acc = (classes_pred == classes_true).sum().item() / len(
                    classes_true)

                # updating confusion matrix
                # noinspection PyUnboundLocalVariable
                confusion_matrix.add(classes_pred.int(), classes_true)
            else:
                # adding the current (y_pred, y_true) couple to the R2Score
                # instance
                # noinspection PyUnboundLocalVariable
                r2_score.update(output=(y_pred, y_true))

            # backward + optimize if training
            if mode == 0:
                # back propagating loss
                batch_loss.backward()

                # making a step with the optimizer
                self.experiment.optimizer.step()

            # printing batch information
            if self.experiment.verbose >= 2:
                # noinspection PyTypeChecker
                expression = ">> {} Batch {}/{} | loss: {:.4f}".format(
                    mode_str,
                    str(batch_index + 1).zfill(len(str(n_batches))),
                    n_batches,
                    float(batch_loss)
                )
                if self.experiment.classification_flag:
                    # noinspection PyUnboundLocalVariable
                    expression += ", acc: {:.4f}".format(batch_acc)
                self.experiment.logger.print(expression)

        # ALL BATCHES DONE!!

        # running lr_scheduler if in validation
        if mode == 1:
            if type(self.experiment.lr_scheduler).__name__ == \
                    "ReduceLROnPlateau":
                self.experiment.lr_scheduler.step(np.mean(batch_loss_list))
            else:
                self.experiment.lr_scheduler.step()

        # getting lr
        lr = None
        for param_group in self.experiment.optimizer.param_groups:
            lr = param_group['lr']

        if self.experiment.classification_flag:
            # computing confusion matrix and accuracy
            cm = confusion_matrix.conf
            acc = sum(cm.diagonal()) / np.sum(cm)
        else:
            # computing r2 score
            r2 = r2_score.compute()

        # computing loss
        loss = float(np.mean(batch_loss_list))

        # filling result dictionaries
        if mode == 0:  # training
            self.train_results["loss_list"].append(loss)
            self.train_results["last_loss"] = loss
            if self.experiment.classification_flag:
                # noinspection PyUnboundLocalVariable
                self.train_results["acc_list"].append(acc)
                self.train_results["last_acc"] = acc
                # noinspection PyUnboundLocalVariable
                self.train_results["cm_list"].append(cm)
            else:
                # noinspection PyUnboundLocalVariable
                self.train_results["r2_list"].append(r2)
                self.train_results["last_r2"] = r2
            #
        elif mode == 1:  # validation
            self.valid_results["loss_list"].append(loss)
            self.valid_results["last_loss"] = loss
            self.valid_results["lr_list"].append(lr)
            if self.experiment.classification_flag:
                # noinspection PyUnboundLocalVariable
                self.valid_results["acc_list"].append(acc)
                self.valid_results["last_acc"] = acc
                # noinspection PyUnboundLocalVariable
                self.valid_results["cm_list"].append(cm)
            else:
                # noinspection PyUnboundLocalVariable
                self.valid_results["r2_list"].append(r2)
                self.valid_results["last_r2"] = r2
            #
        elif mode == 2:  # testing
            # noinspection PyTypeChecker
            self.test_results["loss"] = loss
            if self.experiment.classification_flag:
                # noinspection PyUnboundLocalVariable
                self.test_results["acc"] = acc
                # noinspection PyUnboundLocalVariable
                self.test_results["cm"] = cm
            else:
                # noinspection PyUnboundLocalVariable
                self.test_results["r2"] = r2

    def test(self):
        # if there was validation, loading the best model
        if self.experiment.n_batches_valid > 0:
            self.experiment.model.load_state_dict(
                torch.load(
                    os.path.join(self.experiment.this_dir, 'Net_best_loss.pt')
                )
            )
        self._loop(mode=2)
        if self.experiment.verbose >= 1:
            # noinspection PyStringFormat
            expression = "loss: {:-4f}\n".format(self.test_results['loss'])
            if self.experiment.classification_flag:
                # noinspection PyStringFormat
                expression += "accuracy: {:.4f}\n".format(
                    self.test_results['acc'])
                expression += f"conf matrix:\n{self.test_results['cm']}"
            else:
                # noinspection PyStringFormat
                expression += "r2 score: {:.4f}".format(
                    self.test_results['r2'])
            self.experiment.logger.print(
                expression,
                up_extra_lines=1,
                timestamp_flag=False
            )


if __name__ == '__main__':
    # importing MNIST loader and model
    from pydaves.torchsystem.torchloaders import mnist
    from pydaves.torchsystem.torchmodels import Net3Conv1MP2Lin

    # creating a TorchConfig instance
    config = TorchConfig("config_example.yml")

    # creating a TorchExperiment instance
    experiment = TorchExperiment(
        config=config,
        loader_fun=mnist,
        model_class=Net3Conv1MP2Lin
    )

    # running the experiment
    experiment.run()



"""
TorchSystem.models

This module contains Neural Network classes.
"""

import torch.nn as nn
from pydaves.torchsystem.torchutils import init_params
from pydaves.torchsystem.torchutils import count_maxpool
from pydaves.torchsystem.torchsystem import TorchExperiment


class Net3Conv1MP2Lin(nn.Module):
    """
    This is a simple Neural Networks with 3 Convolutional layers,
    1 Max Pooling layer and 2 liner layers at the end.
    """

    def __init__(self, experiment: TorchExperiment):
        # calling the superclass init method
        super(Net3Conv1MP2Lin, self).__init__()

        # hard copying other useful parameters
        self.activation = experiment.activation
        self.input_height = experiment.input_height
        self.input_width = experiment.input_width
        self.input_depth = experiment.input_depth
        self.n_channels = self.input_depth
        self.n_filters = experiment.model_settings["n_filters"]
        self.pool_size = experiment.model_settings["pool_size"]
        self.kernel_size = experiment.model_settings["kernel_size"]
        self.n_classes = experiment.n_classes
        self.first_linear_layer_out_size = 50

        # creation of the convolutional part of the neural network
        self.convolutional_layers = nn.Sequential(
            # first convolutional layer
            nn.Conv2d(
                in_channels=self.n_channels,
                out_channels=self.n_channels,
                kernel_size=self.kernel_size,
                padding=(1, 1)
            ),

            # activation for the first convolutional layer
            self.activation,

            # first Max Pooling
            nn.MaxPool2d(
                kernel_size=(self.pool_size, self.pool_size),
                stride=(self.pool_size, self.pool_size)
            ),

            # second convolutional layer
            nn.Conv2d(
                self.n_channels,
                self.n_filters,
                kernel_size=self.kernel_size,
                padding=(1, 1)
            ),

            # activation for the second convolutional layer
            self.activation,

            # third convolutional layer
            nn.Conv2d(
                self.n_filters,
                self.n_filters * 2,
                kernel_size=self.kernel_size,
                padding=(1, 1)
            ),

            # activation for the third convolutional layer
            self.activation
        )

        # counting how many pooling layers
        n_pools = count_maxpool(self.convolutional_layers)

        # computing output size
        height_out = self.input_height // (self.pool_size ** n_pools)
        width_out = self.input_width // (self.pool_size ** n_pools)

        # computing linear layer length
        self.first_linear_layer_in_size = height_out * width_out * \
                                          (self.n_filters * 2)

        # creating the first linear layer;
        # it will reduce the dimensionality to 50
        self.classifier0 = nn.Linear(self.first_linear_layer_in_size,
                                     self.first_linear_layer_out_size)

        # crating the output
        self.classifier = nn.Sequential(
            # activation for the first linear layer
            self.activation,

            # creating the second linear layer
            nn.Linear(self.first_linear_layer_out_size, self.n_classes))

        # Initializing all parameters of the Neural Network using the
        # nn.Module.apply method. This method applies a function recursively
        # to every submodule (as returned by .children()) as well as self.
        self.apply(init_params)

    def forward(self, x):
        # computing the output for the convolutional part of the Neural Network
        x = self.convolutional_layers(x)

        # The view method returns a new tensor with the same data as the self
        # tensor but of a different shape. The size -1 is inferred from
        # other dimensions.
        x = x.view(-1, self.first_linear_layer_in_size)

        # computing the output from the first linear layer
        x0 = self.classifier0(x)

        # computing the final output
        x = self.classifier(x0)

        # returning the final output
        return x
